#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional

from jdmn.feel.lib.type.EqualityComparator import EqualityComparator
from jdmn.feel.lib.type.bool.TernaryBooleanLogicUtil import TernaryBooleanLogicUtil


class RelationalComparator(EqualityComparator):
    def __init__(self):
        super().__init__()

    def compare(self, first: Any, second: Any) -> int:
        pass

    def lessThan(self, first: Any, second: Any) -> Optional[bool]:
        pass

    def greaterThan(self, first: Any, second: Any) -> Optional[bool]:
        return self.lessThan(second, first)

    def lessEqualThan(self, first: Any, second: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().or_(self.lessThan(first, second), self.equalTo(first, second))

    def greaterEqualThan(self, first: Any, second: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().or_(self.greaterThan(first, second), self.equalTo(first, second))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License")or you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.feel.lib.FEELLib import FEELLib
from jdmn.feel.lib.Types import DECIMAL, BOOLEAN, DATE, TIME, DATE_TIME, DURATION, LIST, INTEGER, STRING, CONTEXT, RANGE
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType
from jdmn.feel.lib.type.context.DefaultContextType import DefaultContextType
from jdmn.feel.lib.type.function.DefaultFunctionType import DefaultFunctionType
from jdmn.feel.lib.type.list.DefaultListType import DefaultListType
from jdmn.feel.lib.type.numeric.DefaultNumericType import DefaultNumericType
from jdmn.feel.lib.type.range.DefaultRangeType import DefaultRangeType
from jdmn.feel.lib.type.string.DefaultStringType import DefaultStringType
from jdmn.feel.lib.type.time.DefaultDateTimeType import DefaultDateTimeType
from jdmn.feel.lib.type.time.DefaultDateType import DefaultDateType
from jdmn.feel.lib.type.time.DefaultDurationType import DefaultDurationType
from jdmn.feel.lib.type.time.DefaultTimeType import DefaultTimeType
from jdmn.runtime.LazyEval import LazyEval
from jdmn.runtime.listener.EventListener import EventListener
from jdmn.runtime.listener.Rule import Rule


class BaseFEELLib(FEELLib):
    # Types
    numericType = DefaultNumericType()
    booleanType = DefaultBooleanType()
    stringType = DefaultStringType()
    dateType = DefaultDateType()
    timeType = DefaultTimeType()
    dateTimeType = DefaultDateTimeType()
    durationType = DefaultDurationType()
    listType = DefaultListType()
    contextType = DefaultContextType()
    rangeType = DefaultRangeType()
    functionType = DefaultFunctionType()

    def __init__(self):
        FEELLib.__init__(self)

    #
    # Numeric operators
    #
    def isNumber(self, value: Any) -> bool:
        try:
            return self.numericType.isNumber(value)
        except Exception as e:
            message = "isNumber({})".format(value)
            self.logError(message, e)
            return False

    def numericValue(self, value: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericValue(value)
        except Exception as e:
            message = "numericValue({})".format(value)
            self.logError(message, e)
            return None

    def numericIs(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericIs(first, second)
        except Exception as e:
            message = "numericIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericEqual(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericEqual(first, second)
        except Exception as e:
            message = "numericEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericNotEqual(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericNotEqual(first, second)
        except Exception as e:
            message = "numericNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericLessThan(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericLessThan(first, second)
        except Exception as e:
            message = "numericLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericGreaterThan(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericGreaterThan(first, second)
        except Exception as e:
            message = "numericGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericLessEqualThan(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericLessEqualThan(first, second)
        except Exception as e:
            message = "numericLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericGreaterEqualThan(self, first: DECIMAL, second: DECIMAL) -> BOOLEAN:
        try:
            return self.numericType.numericGreaterEqualThan(first, second)
        except Exception as e:
            message = "numericGreaterEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericAdd(self, first: DECIMAL, second: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericAdd(first, second)
        except Exception as e:
            message = "numericAdd({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericSubtract(self, first: DECIMAL, second: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericSubtract(first, second)
        except Exception as e:
            message = "numericSubtract({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericMultiply(self, first: DECIMAL, second: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericMultiply(first, second)
        except Exception as e:
            message = "numericMultiply({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericDivide(self, first: DECIMAL, second: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericDivide(first, second)
        except Exception as e:
            message = "numericDivide({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def numericUnaryMinus(self, first: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericUnaryMinus(first)
        except Exception as e:
            message = "numericUnaryMinus({}".format(first)
            self.logError(message, e)
            return None

    def numericExponentiation(self, first: DECIMAL, second: DECIMAL) -> DECIMAL:
        try:
            return self.numericType.numericExponentiation(first, second)
        except Exception as e:
            message = "numericExponentiation({}, {})".format(first, second)
            self.logError(message, e)
            return None

    #
    # Boolean operators
    #
    def isBoolean(self, value: Any) -> BOOLEAN:
        try:
            return self.booleanType.isBoolean(value)
        except Exception as e:
            message = "isBoolean({})".format(value)
            self.logError(message, e)
            return False

    def booleanValue(self, value: BOOLEAN) -> BOOLEAN:
        try:
            return self.booleanType.booleanValue(value)
        except Exception as e:
            message = "booleanValue({})".format(value)
            self.logError(message, e)
            return None

    def booleanIs(self, first: BOOLEAN, second: BOOLEAN) -> BOOLEAN:
        try:
            return self.booleanType.booleanIs(first, second)
        except Exception as e:
            message = "booleanIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def booleanEqual(self, first: BOOLEAN, second: BOOLEAN) -> BOOLEAN:
        try:
            return self.booleanType.booleanEqual(first, second)
        except Exception as e:
            message = "booleanEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def booleanNotEqual(self, first: BOOLEAN, second: BOOLEAN) -> BOOLEAN:
        try:
            return self.booleanType.booleanNotEqual(first, second)
        except Exception as e:
            message = "booleanNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def booleanNot(self, operand: Any) -> BOOLEAN:
        try:
            return self.booleanType.booleanNot(operand)
        except Exception as e:
            message = "booleanNot({})".format(operand)
            self.logError(message, e)
            return None

    def booleanOr(self, *operands) -> BOOLEAN:
        try:
            return self.booleanType.booleanOr(*operands)
        except Exception as e:
            message = "booleanOr({})".format(operands)
            self.logError(message, e)
            return None

    def binaryBooleanOr(self, first: Any, second: Any) -> BOOLEAN:
        try:
            return self.booleanType.binaryBooleanOr(first, second)
        except Exception as e:
            message = "binaryBooleanOr({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def booleanAnd(self, *operands) -> BOOLEAN:
        try:
            return self.booleanType.booleanAnd(*operands)
        except Exception as e:
            message = "booleanAnd({})".format(operands)
            self.logError(message, e)
            return None

    def binaryBooleanAnd(self, first: Any, second: Any) -> BOOLEAN:
        try:
            return self.booleanType.binaryBooleanAnd(first, second)
        except Exception as e:
            message = "binaryBooleanAnd({}, {})".format(first, second)
            self.logError(message, e)
            return None

    #
    # String operators
    #
    def isString(self, value: Any) -> BOOLEAN:
        try:
            return self.stringType.isString(value)
        except Exception as e:
            message = "isString({})".format(value)
            self.logError(message, e)
            return False

    def stringValue(self, value: STRING) -> STRING:
        try:
            return self.stringType.stringValue(value)
        except Exception as e:
            message = "stringValue({})".format(value)
            self.logError(message, e)
            return None

    def stringIs(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringIs(first, second)
        except Exception as e:
            message = "stringIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringEqual(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringEqual(first, second)
        except Exception as e:
            message = "stringEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringNotEqual(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringNotEqual(first, second)
        except Exception as e:
            message = "stringNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringLessThan(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringLessThan(first, second)
        except Exception as e:
            message = "stringLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringGreaterThan(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringGreaterThan(first, second)
        except Exception as e:
            message = "stringGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringLessEqualThan(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringLessEqualThan(first, second)
        except Exception as e:
            message = "stringLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringGreaterEqualThan(self, first: STRING, second: STRING) -> BOOLEAN:
        try:
            return self.stringType.stringGreaterEqualThan(first, second)
        except Exception as e:
            message = "({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def stringAdd(self, first: STRING, second: STRING) -> STRING:
        try:
            return self.stringType.stringAdd(first, second)
        except Exception as e:
            message = "stringAdd({}, {})".format(first, second)
            self.logError(message, e)
            return None

    #
    # Date operators
    #
    def isDate(self, value: Any) -> BOOLEAN:
        try:
            return self.dateType.isDate(value)
        except Exception as e:
            message = "isDate({})".format(value)
            self.logError(message, e)
            return False

    def dateValue(self, date: DATE) -> INTEGER:
        try:
            return self.dateType.dateValue(date)
        except Exception as e:
            message = "dateValue({})".format(date)
            self.logError(message, e)
            return None

    def dateIs(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateIs(first, second)
        except Exception as e:
            message = "dateIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateEqual(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateEqual(first, second)
        except Exception as e:
            message = "dateEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateNotEqual(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateNotEqual(first, second)
        except Exception as e:
            message = "dateNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateLessThan(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateLessThan(first, second)
        except Exception as e:
            message = "dateLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateGreaterThan(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateGreaterThan(first, second)
        except Exception as e:
            message = "dateGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateLessEqualThan(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateLessEqualThan(first, second)
        except Exception as e:
            message = "dateLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateGreaterEqualThan(self, first: DATE, second: DATE) -> BOOLEAN:
        try:
            return self.dateType.dateGreaterEqualThan(first, second)
        except Exception as e:
            message = "dateGreaterEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateSubtract(self, first: DATE, second: DATE) -> DURATION:
        try:
            dateTime1: DATE_TIME = self.toDateTime(first)
            dateTime2: DATE_TIME = self.toDateTime(second)
            return self.dateTimeType.dateTimeSubtract(dateTime1, dateTime2)
        except Exception as e:
            message = "dateSubtract({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateAddDuration(self, date: DATE, duration: DURATION) -> DATE:
        try:
            return self.dateType.dateAddDuration(date, duration)
        except Exception as e:
            message = "dateAddDuration({}, {})".format(date, duration)
            self.logError(message, e)
            return None

    def dateSubtractDuration(self, date: DATE, duration: DURATION) -> DATE:
        try:
            return self.dateType.dateSubtractDuration(date, duration)
        except Exception as e:
            message = "dateSubtractDuration({}, {})".format(date, duration)
            self.logError(message, e)
            return None

    #
    # Time operators
    #
    def isTime(self, value: Any) -> BOOLEAN:
        try:
            return self.timeType.isTime(value)
        except Exception as e:
            message = "isTime({})".format(value)
            self.logError(message, e)
            return False

    def timeValue(self, time: TIME) -> INTEGER:
        try:
            return self.timeType.timeValue(time)
        except Exception as e:
            message = "timeValue({})".format(time)
            self.logError(message, e)
            return None

    def timeIs(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeIs(first, second)
        except Exception as e:
            message = "timeIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeEqual(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeEqual(first, second)
        except Exception as e:
            message = "timeEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeNotEqual(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeNotEqual(first, second)
        except Exception as e:
            message = "timeNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeLessThan(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeLessThan(first, second)
        except Exception as e:
            message = "timeLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeGreaterThan(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeGreaterThan(first, second)
        except Exception as e:
            message = "timeGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeLessEqualThan(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeLessEqualThan(first, second)
        except Exception as e:
            message = "timeLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeGreaterEqualThan(self, first: TIME, second: TIME) -> BOOLEAN:
        try:
            return self.timeType.timeGreaterEqualThan(first, second)
        except Exception as e:
            message = "timeGreaterEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeSubtract(self, first: TIME, second: TIME) -> DURATION:
        try:
            return self.timeType.timeSubtract(first, second)
        except Exception as e:
            message = "timeSubtract({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def timeAddDuration(self, time: TIME, duration: DURATION) -> TIME:
        try:
            return self.timeType.timeAddDuration(time, duration)
        except Exception as e:
            message = "timeAddDuration({}, {})".format(time, duration)
            self.logError(message, e)
            return None

    def timeSubtractDuration(self, time: TIME, duration: DURATION) -> TIME:
        try:
            return self.timeType.timeSubtractDuration(time, duration)
        except Exception as e:
            message = "timeSubtractDuration({}, {})".format(time, duration)
            self.logError(message, e)
            return None

    #
    # Date and Time operators
    #
    def isDateTime(self, value: Any) -> BOOLEAN:
        try:
            return self.dateTimeType.isDateTime(value)
        except Exception as e:
            message = "isDateTime({})".format(value)
            self.logError(message, e)
            return False

    def dateTimeValue(self, dateTime: DATE_TIME) -> INTEGER:
        try:
            return self.dateTimeType.dateTimeValue(dateTime)
        except Exception as e:
            message = "dateTimeValue({})".format(dateTime)
            self.logError(message, e)
            return None

    def dateTimeIs(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeIs(first, second)
        except Exception as e:
            message = "dateTimeIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeEqual(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeEqual(first, second)
        except Exception as e:
            message = "dateTimeEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeNotEqual(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeNotEqual(first, second)
        except Exception as e:
            message = "dateTimeNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeLessThan(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeLessThan(first, second)
        except Exception as e:
            message = "dateTimeLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeGreaterThan(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeGreaterThan(first, second)
        except Exception as e:
            message = "dateTimeGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeLessEqualThan(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeLessEqualThan(first, second)
        except Exception as e:
            message = "dateTimeLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeGreaterEqualThan(self, first: DATE_TIME, second: DATE_TIME) -> BOOLEAN:
        try:
            return self.dateTimeType.dateTimeGreaterEqualThan(first, second)
        except Exception as e:
            message = "dateTimeGreaterEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeSubtract(self, first: DATE_TIME, second: DATE_TIME) -> DURATION:
        try:
            if self.isDate(first):
                first = self.toDateTime(first)
            if self.isDate(second):
                second = self.toDateTime(second)

            return self.dateTimeType.dateTimeSubtract(first, second)
        except Exception as e:
            message = "dateTimeSubtract({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def dateTimeAddDuration(self, dateTime: DATE_TIME, duration: DURATION) -> DATE_TIME:
        try:
            return self.dateTimeType.dateTimeAddDuration(dateTime, duration)
        except Exception as e:
            message = "dateTimeAddDuration({}, {})".format(dateTime, duration)
            self.logError(message, e)
            return None

    def dateTimeSubtractDuration(self, dateTime: DATE_TIME, duration: DURATION) -> DATE_TIME:
        try:
            return self.dateTimeType.dateTimeSubtractDuration(dateTime, duration)
        except Exception as e:
            message = "dateTimeSubtractDuration({}, {})".format(dateTime, duration)
            self.logError(message, e)
            return None

    #
    # Duration operators
    #
    def isDuration(self, value: Any) -> BOOLEAN:
        try:
            return self.durationType.isDuration(value)
        except Exception as e:
            message = "isDuration({})".format(value)
            self.logError(message, e)
            return False

    def isYearsAndMonthsDuration(self, value: Any) -> BOOLEAN:
        try:
            return self.durationType.isYearsAndMonthsDuration(value)
        except Exception as e:
            message = "isYearsAndMonthsDuration({})".format(value)
            self.logError(message, e)
            return False

    def isDaysAndTimeDuration(self, value: Any) -> BOOLEAN:
        try:
            return self.durationType.isDaysAndTimeDuration(value)
        except Exception as e:
            message = "isDaysAndTimeDuration({})".format(value)
            self.logError(message, e)
            return False

    def durationIs(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationIs(first, second)
        except Exception as e:
            message = "durationIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationValue(self, duration: DURATION) -> INTEGER:
        try:
            return self.durationType.durationValue(duration)
        except Exception as e:
            message = "durationValue({})".format(duration)
            self.logError(message, e)
            return None

    def durationEqual(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationEqual(first, second)
        except Exception as e:
            message = "durationEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationNotEqual(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationNotEqual(first, second)
        except Exception as e:
            message = "durationNotEqual({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationLessThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationLessThan(first, second)
        except Exception as e:
            message = "durationLessThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationGreaterThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationGreaterThan(first, second)
        except Exception as e:
            message = "durationGreaterThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationLessEqualThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationLessEqualThan(first, second)
        except Exception as e:
            message = "durationLessEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationGreaterEqualThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        try:
            return self.durationType.durationGreaterEqualThan(first, second)
        except Exception as e:
            message = "durationGreaterEqualThan({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationAdd(self, first: DURATION, second: DURATION) -> DURATION:
        try:
            return self.durationType.durationAdd(first, second)
        except Exception as e:
            message = "durationAdd({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationSubtract(self, first: DURATION, second: DURATION) -> DURATION:
        try:
            return self.durationType.durationSubtract(first, second)
        except Exception as e:
            message = "durationSubtract({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationDivide(self, first: DURATION, second: DURATION) -> DECIMAL:
        try:
            return self.durationType.durationDivide(first, second)
        except Exception as e:
            message = "durationMultiply({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationMultiplyNumber(self, first: DURATION, second: DECIMAL) -> DURATION:
        try:
            return self.durationType.durationMultiplyNumber(first, second)
        except Exception as e:
            message = "durationMultiply({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def durationDivideNumber(self, first: DURATION, second: DECIMAL) -> DURATION:
        try:
            return self.durationType.durationDivideNumber(first, second)
        except Exception as e:
            message = "durationDivide({}, {})".format(first, second)
            self.logError(message, e)
            return None

    #
    # List operators
    #
    def isList(self, value: Any) -> BOOLEAN:
        try:
            return self.listType.isList(value)
        except Exception as e:
            message = "isList({})".format(value)
            self.logError(message, e)
            return False

    def listValue(self, value: LIST) -> LIST:
        try:
            return self.listType.listValue(value)
        except Exception as e:
            message = "listValue({})".format(value)
            self.logError(message, e)
            return None

    def listIs(self, first: LIST, second: LIST) -> BOOLEAN:
        try:
            return self.listType.listIs(first, second)
        except Exception as e:
            message = "listIs({}, {})".format(first, second)
            self.logError(message, e)
            return None

    def listEqual(self, list1: LIST, list2: LIST) -> BOOLEAN:
        try:
            return self.listType.listEqual(list1, list2)
        except Exception as e:
            message = "listEqual({}, {})".format(list1, list2)
            self.logError(message, e)
            return None

    def listNotEqual(self, list1: LIST, list2: LIST) -> BOOLEAN:
        try:
            return self.listType.listNotEqual(list1, list2)
        except Exception as e:
            message = "listNotEqual({}, {})".format(list1, list2)
            self.logError(message, e)
            return None

    #
    # Context operators
    #
    def isContext(self, value: Any) -> BOOLEAN:
        try:
            return self.contextType.isContext(value)
        except Exception as e:
            message = "isContext({})".format(value)
            self.logError(message, e)
            return False

    def contextValue(self, value: CONTEXT) -> CONTEXT:
        try:
            return self.contextType.contextValue(value)
        except Exception as e:
            message = "contextValue({})".format(value)
            self.logError(message, e)
            return None

    def contextIs(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        try:
            return self.contextType.contextIs(c1, c2)
        except Exception as e:
            message = "contextIs({}, {})".format(c1, c2)
            self.logError(message, e)
            return None

    def contextEqual(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        try:
            return self.contextType.contextEqual(c1, c2)
        except Exception as e:
            message = "contextEqual({}, {})".format(c1, c2)
            self.logError(message, e)
            return None

    def contextNotEqual(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        try:
            return self.contextType.contextNotEqual(c1, c2)
        except Exception as e:
            message = "contextNotEqual({}, {})".format(c1, c2)
            self.logError(message, e)
            return None

    #
    # Context functions
    #
    def getEntries(self, m: CONTEXT) -> LIST:
        try:
            return self.contextType.getEntries(m)
        except Exception as e:
            message = "getEntries({})".format(m)
            self.logError(message, e)
            return None

    def getValue(self, m: CONTEXT, key: Any) -> Any:
        try:
            return self.contextType.getValue(m, key)
        except Exception as e:
            message = "getValue({}, {})".format(m, key)
            self.logError(message, e)
            return None

    #
    # Range operators
    #
    def isRange(self, value: Any) -> BOOLEAN:
        try:
            return self.rangeType.isRange(value)
        except Exception as e:
            message = "isRange({})".format(value)
            self.logError(message, e)
            return False

    def rangeValue(self, value: RANGE) -> RANGE:
        try:
            return self.rangeType.rangeValue(value)
        except Exception as e:
            message = "rangeValue({})".format(value)
            self.logError(message, e)
            return None

    def rangeIs(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        try:
            return self.rangeType.rangeIs(range1, range2)
        except Exception as e:
            message = "rangeIs({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def rangeEqual(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        try:
            return self.rangeType.rangeEqual(range1, range2)
        except Exception as e:
            message = "rangeEqual({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def rangeNotEqual(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        try:
            return self.rangeType.rangeNotEqual(range1, range2)
        except Exception as e:
            message = "rangeNotEqual({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    #
    # Function operators
    #
    def isFunction(self, value: Any) -> BOOLEAN:
        try:
            return self.functionType.isFunction(value)
        except Exception as e:
            message = "isFunction({})".format(value)
            self.logError(message, e)
            return False

    def functionValue(self, value: Any) -> Any:
        try:
            return self.functionType.functionValue(value)
        except Exception as e:
            message = "functionValue({})".format(value)
            self.logError(message, e)
            return None

    def functionIs(self, function1: Any, function2: Any) -> BOOLEAN:
        try:
            return self.functionType.functionIs(function1, function2)
        except Exception as e:
            message = "functionIs({}, {})".format(function1, function2)
            self.logError(message, e)
            return None

    def functionEqual(self, function1: Any, function2: Any) -> BOOLEAN:
        try:
            return self.functionType.functionEqual(function1, function2)
        except Exception as e:
            message = "functionEqual({}, {})".format(function1, function2)
            self.logError(message, e)
            return None

    def functionNotEqual(self, function1: Any, function2: Any) -> BOOLEAN:
        try:
            return self.functionType.functionNotEqual(function1, function2)
        except Exception as e:
            message = "functionNotEqual({}, {})".format(function1, function2)
            self.logError(message, e)
            return None

    #
    # Conversion functions
    #
    def asList(self, *args) -> LIST:
        if len(args) == 0:
            return []
        else:
            return list(args)

    def asElement(self, list_: LIST) -> Any:
        if list_ is None:
            return None
        elif len(list_) == 1:
            return list_[0]
        else:
            return None

    #
    # Other functions
    #
    def rangeToList(self, *args) -> LIST:
        len_ = len(args)
        if len_ == 2:
            arg1 = args[0]
            arg2 = args[1]
            return self.rangeToListWithArgs(False, arg1, False, arg2)
        elif len_ == 4:
            return self.rangeToListWithArgs(args[0], args[1], args[2], args[3])
        else:
            return None

    def rangeToListWithArgs(self, isOpenStart: bool, start: DECIMAL, isOpenEnd: bool, end: DECIMAL) -> LIST:
        if start is None or end is None:
            return []
        else:
            range_ = self.intRange(isOpenStart, start, isOpenEnd, end)
            return [x for x in range_]

    def intRange(self, isOpenStart: bool, start: DECIMAL, isOpenEnd: bool, end: DECIMAL):
        if start is None or end is None:
            return None

        startValue: int = int(start) + 1 if isOpenStart else int(start)
        endValue: int = int(end) - 1 if isOpenEnd else int(end)
        if startValue <= endValue:
            return range(startValue, endValue + 1)
        else:
            return range(startValue, endValue - 1, -1)

    def flattenFirstLevel(self, list_: LIST) -> LIST:
        if list_ is None:
            return None

        result = []
        for obj in list_:
            if isinstance(obj, list):
                result.extend(obj)
            else:
                result.append(obj)

        return result

    def elementAt(self, list: LIST, number: DECIMAL) -> Any:
        if list is None:
            return None

        index = int(number)

        listSize = len(list)
        if 1 <= index and index <= listSize:
            return list[index - 1]
        elif -listSize <= index and index <= -1:
            return list[listSize + index]
        else:
            return None

    def ruleMatches(self, eventListener: EventListener, rule: Rule, *operands) -> bool:
        if (operands is None or len(operands) == 0):
            return False
        else:
            for i, operand in enumerate(operands):
                if isinstance(operand, LazyEval):
                    operand = operand.getOrCompute()

                # Column index starts first 1
                eventListener.matchColumn(rule, i + 1, operand)

                if operand is None or operand is False:
                    return False
            return True

    def valueOf(self, number: int) -> DECIMAL:
        pass

    def intValue(self, number: DECIMAL) -> int:
        pass



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any


class Context:
    def __init__(self, name: str = None):
        self.name = name
        self.map = {}

    def getName(self) -> str:
        return self.name

    def getBindings(self) -> dict:
        return self.map

    def get(self, name: str, *args) -> Any:
        o = self.map[name]
        if o is not None:
            return o
        for key in args:
            o = self.map[key]
            if o is not None:
                return o
        return None

    def put(self, key: Any, value: Any) -> Any:
        self.map[key] = value
        return value

    def add(self, key: Any, value: Any):
        self.put(key, value)
        return self

    def isEquivalent(self, other: Any):
        return other is not None and isinstance(other, Context) and self.keySet() == other.keySet()

    def __str__(self):
        return str(self.map)

    def keySet(self):
        return self.map.keys()

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        if other is None or self.__class__ != other.__class__:
            return False

        return self.map == other.map



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import List, Any


class ExternalFunctionExecutor:
    def execute(self, className: str, methodName: str, args: List[Any]) -> Any:
        pass



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any


class Range:
    def __init__(self, startIncluded: bool, start: Any, endIncluded: bool, end: Any):
        self.startIncluded = startIncluded
        self.start = start
        self.endIncluded = endIncluded
        self.end = end

    def isStartIncluded(self) -> bool:
        return self.startIncluded

    def getStart(self) -> Any:
        return self.start

    def isEndIncluded(self) -> bool:
        return self.endIncluded

    def getEnd(self) -> Any:
        return self.end

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if type(self) is not type(other):
            return False

        if self.startIncluded != other.startIncluded:
            return False
        if self.endIncluded != other.endIncluded:
            return False
        if self.start != other.start:
            return False
        if self.end != other.end:
            return False
        return True

    def __hash__(self):
        result = 0
        result = 31 * result + (0 if self.startIncluded is None else hash(self.startIncluded))
        result = 31 * result + (0 if self.start is None else hash(self.start))
        result = 31 * result + (0 if self.endIncluded is None else hash(self.endIncluded))
        result = 31 * result + (0 if self.end is None else hash(self.end))
        return result

    def __str__(self):
        return "Range({},{},{},{})".format(self.startIncluded, self.start, self.end, self.endIncluded)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import re
from datetime import date, time, datetime
from decimal import Decimal
from typing import Any
from zoneinfo import ZoneInfo

import isodate
from isodate import Duration, tzinfo, FixedOffset

from jdmn.feel.lib.Types import STRING, DATE, TIME, DATE_TIME, DURATION, INTEGER, TIME_OR_DATE_TIME
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.runtime.DMNRuntimeException import DMNRuntimeException

DAY_NAMES = ["", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
MONTH_NAMES = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]


class DefaultDateTimeLib:
    DATE_PART = r"\d{4}-\d{2}-\d{2}"
    # [T] HH : MM : SS [.S+] (DDDD | DD:DD | Z )
    TIME_PART = r"\d{2}:\d{2}:\d{2}([.]\d+)?([+-]\d{4}|[+-]\d{2}:\d{2}|[Zz])?"

    DATE_PATTERN = re.compile("^" + DATE_PART + "$")
    TIME_PATTERN = re.compile("^T?" + TIME_PART + "$")
    DATE_TIME_PATTERN = re.compile("^" + DATE_PART + "T" + TIME_PART + "$")

    #
    # Conversion functions
    #
    def date(self, *args) -> DATE:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal
                result = self.parseDate(arg)
            elif isinstance(arg, datetime) or isinstance(arg, date):
                # From date or datetime
                result = self.toDate(arg)
            else:
                result = None
        elif len_ == 3:
            # From year, month, day
            year = int(args[0])
            month = int(args[1])
            day = int(args[2])
            result = date(year, month, day)
        else:
            result = None
        return result if self.isValidDateValue(result) else None

    def time(self, *args) -> TIME:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal
                if '@' in arg:
                    parts = arg.split("@")
                    time_ = self.parseTime(parts[0])
                    result = self.mergeTzInfo(arg, time_, parts[1])
                else:
                    result = self.parseTime(arg)
            elif isinstance(arg, time) or isinstance(arg, datetime) or isinstance(arg, date):
                # From date, time or datetime
                result = self.toTime(arg)
            else:
                result = None
        elif len_ == 4:
            # From hour, minute, second, offset
            hour = int(args[0])
            minute = int(args[1])
            seconds = int(args[2])
            secondFraction = args[2] - seconds
            micros = int(secondFraction * Decimal(1e6))
            offset = self.toTzinfo(args[3])
            if offset is None:
                result = time(hour=hour, minute=minute, second=seconds, microsecond=micros)
            else:
                result = time(hour=hour, minute=minute, second=seconds, microsecond=micros, tzinfo=offset)
        else:
            result = None
        return result if self.isValidTimeValue(result) else None

    def dateAndTime(self, *args) -> DATE_TIME:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal (date or datetime)
                if "-" in arg and not ("T" in arg):
                    arg += "T00:00:00"
                if '@' in arg:
                    parts = arg.split("@")
                    dt = self.parseDateTime(parts[0])
                    result = self.mergeTzInfo(arg, dt, parts[1])
                else:
                    result = self.parseDateTime(arg)
            elif isinstance(arg, time):
                # From time
                result = arg
            else:
                result = None
        elif len_ == 2:
            # From date, time
            result = datetime.combine(args[0], args[1])
        else:
            result = None
        return result if self.isValidDateTimeValue(result) else None

    def parseDate(self, arg: str) -> date:
        if not bool(self.DATE_PATTERN.match(arg)):
            raise DMNRuntimeException("Illegal date format '{}'".format(arg))

        return date.fromisoformat(arg)

    def parseTime(self, arg: str) -> time:
        if not bool(self.TIME_PATTERN.match(arg)):
            raise DMNRuntimeException("Illegal time format '{}'".format(arg))

        arg = self.fixDateTimeFormat(arg)
        #        t = time.fromisoformat(arg)
        t = isodate.parse_time(arg)
        #        t = dateutil.parser.isoparse(arg)
        return t

    def parseDateTime(self, arg: str) -> datetime:
        if not bool(self.DATE_TIME_PATTERN.match(arg)):
            raise DMNRuntimeException("Illegal datetime format '{}'".format(arg))

        arg = self.fixDateTimeFormat(arg)
        #        dt = time.fromisoformat(arg)
        try:
            datestring, timestring = arg.split('T')
        except ValueError:
            raise DMNRuntimeException("ISO 8601 time designator 'T' missing. Unable to parse datetime string {}".format(arg))

        tmpdate = isodate.parse_date(datestring, defaultday=0, defaultmonth=0)
        tmptime = isodate.parse_time(timestring)
        dt = datetime.combine(tmpdate, tmptime)
        #        dt = dateutil.parser.isoparse(arg)
        return dt

    # Fix the format 2016-08-01T11:00:00.000+0000 to 2016-08-01T11:00:00.000+00:00
    # and T11:00:00.000+0000 to 11:00:00.000+00:00
    def fixDateTimeFormat(self, literal: STRING) -> STRING:
        if literal is None:
            return None
        if (literal.startswith("T")):
            literal = literal[1:]
        timeZoneStartIndex = len(literal) - 5
        if 0 <= timeZoneStartIndex < len(literal):
            timeZoneStart = literal[timeZoneStartIndex]
            if timeZoneStart == '+' or timeZoneStart == '-':
                timeZoneOffset = literal[timeZoneStartIndex + 1:]
                literal = literal[0:timeZoneStartIndex + 1] + timeZoneOffset[0:2] + ":" + timeZoneOffset[2:]
        # Python does not support Z or z as tzinfo
        literal = literal.upper()
        literal = literal.replace("Z", "+00:00")
        return literal

    def mergeTzInfo(self, arg, t: TIME_OR_DATE_TIME, zoneId: str) -> TIME_OR_DATE_TIME:
        if t.tzinfo is None:
            tz = ZoneInfo(zoneId)
            if tz is None:
                raise DMNRuntimeException("Illegal timezone '{}'".format(zoneId))

            result = t.replace(tzinfo=tz)
            return result
        else:
            if not (isinstance(time.tzinfo, isodate.tzinfo.Utc) and zoneId == "UTC"):
                raise DMNRuntimeException("Duplicated timezone in '{}'".format(arg))
            else:
                return t

    #
    # Date properties
    #
    def year(self, date: DATE) -> INTEGER:
        if date is None:
            return None

        return date.year

    def yearDateTime(self, dateTime: DATE_TIME) -> INTEGER:
        return self.year(dateTime)

    def month(self, date: DATE) -> INTEGER:
        if date is None:
            return None

        return date.month

    def monthDateTime(self, dateTime: DATE_TIME) -> INTEGER:
        return self.month(dateTime)

    def day(self, date: DATE) -> INTEGER:
        if date is None:
            return None

        return date.day

    def dayDateTime(self, dateTime: DATE_TIME) -> INTEGER:
        return self.day(dateTime)

    def weekday(self, date: DATE) -> INTEGER:
        if date is None:
            return None

        return date.weekday() + 1

    def weekdayDateTime(self, dateTime: DATE_TIME) -> INTEGER:
        return self.weekday(dateTime)

    #
    # Time properties
    #
    def hour(self, time: TIME_OR_DATE_TIME) -> INTEGER:
        if time is None:
            return None

        return time.hour

    def minute(self, time: TIME_OR_DATE_TIME) -> INTEGER:
        if time is None:
            return None

        return time.minute

    def second(self, time: TIME_OR_DATE_TIME) -> INTEGER:
        if time is None:
            return None

        return time.second

    def timeOffset(self, time: TIME_OR_DATE_TIME) -> DURATION:
        if time is None:
            return None

        return self.toDuration(time.tzinfo)

    def timezone(self, time: TIME_OR_DATE_TIME) -> STRING:
        if time is None:
            return None

        return time.tzname()

    #
    # Temporal functions
    #
    def dayOfYear(self, date: DATE) -> INTEGER:
        if date is None:
            return None

        return date.timetuple().tm_yday

    def dayOfWeek(self, date: DATE) -> STRING:
        if date is None:
            return None

        dow: int = date.isocalendar()[2]
        return DAY_NAMES[dow]

    def weekOfYear(self, date: DATE) -> INTEGER:
        if (date is None):
            return None

        return date.isocalendar()[1]

    def monthOfYear(self, date: DATE) -> STRING:
        if date is None:
            return None

        moy: int = date.month
        return MONTH_NAMES[moy - 1]

    #
    # Extra conversion functions
    #
    def toDate(self, from_: Any) -> DATE:
        if isinstance(from_, datetime):
            return from_.date()
        elif isinstance(from_, date):
            return from_
        else:
            return None

    def toTime(self, from_: Any) -> TIME:
        if isinstance(from_, time):
            return from_
        elif isinstance(from_, datetime):
            return from_.timetz()
        elif isinstance(from_, date):
            dateTime = self.toDateTime(from_)
            return dateTime.timetz()
        else:
            return None

    def toDateTime(self, from_: Any) -> DATE_TIME:
        if isinstance(from_, datetime):
            return from_
        elif isinstance(from_, date):
            return datetime.combine(from_, time(0, 0, 0, tzinfo=tzinfo.UTC))
        else:
            return None

    def toDuration(self, info: tzinfo) -> DURATION:
        if info is None:
            return None

        delta = info.utcoffset(None)
        seconds = int(delta.total_seconds())
        return Duration(seconds=seconds)

    def toTzinfo(self, duration: DURATION) -> tzinfo:
        if duration is None:
            return None

        seconds = duration.total_seconds()
        mm, ss = divmod(seconds, 60)
        hh, mm = divmod(mm, 60)
        if -24 <= hh <= 24:
            return FixedOffset(offset_hours=int(hh), offset_minutes=int(mm))
        else:
            raise DMNRuntimeException("Incorrect offset {}:{}:{}".format(hh, mm, ss))

    def isValidDateValue(self, value: DATE) -> bool:
        return self.isValidDate(value.year, value.month, value.day)

    def isValidTimeValue(self, value: TIME) -> bool:
        secondsOffset = None if value.tzinfo is None or value.utcoffset() is None else value.tzinfo.utcoffset(BaseType.EP).total_seconds()
        return self.isValidTime(value.hour, value.minute, value.second, secondsOffset)

    def isValidDateTimeValue(self, value: DATE_TIME) -> bool:
        secondsOffset = None if value.tzinfo is None or value.utcoffset() is None else value.utcoffset().total_seconds()
        return self.isValidDate(value.year, value.month, value.day) and self.isValidTime(value.hour, value.minute, value.second, secondsOffset)

    def isValidDate(self, year: int, month: int, day: int) -> bool:
        return self.isValidYear(year) and self.isValidMonth(month) and self.isValidDay(day)

    def isValidTime(self, hour: int, minute: int, second: int, secondsOffset: INTEGER) -> bool:
        return self.isValidHour(hour) and self.isValidMinute(minute) and self.isValidSecond(second) and self.isValidOffset(secondsOffset)

    def isValidDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, secondsOffset: INTEGER) -> bool:
        return self.isValidDate(year, month, day) and self.isValidTime(hour, minute, second, secondsOffset)

    @staticmethod
    def isValidYear(year: int) -> bool:
        return -999999999 <= year <= 999999999

    @staticmethod
    def isValidMonth(month: int) -> bool:
        return 1 <= month <= 12

    @staticmethod
    def isValidDay(day: int) -> bool:
        return 1 <= day <= 31

    @staticmethod
    def isValidHour(hour: int) -> bool:
        return 0 <= hour <= 23

    @staticmethod
    def isValidMinute(minute: int) -> bool:
        return 0 <= minute <= 59

    @staticmethod
    def isValidSecond(second: int) -> bool:
        return 0 <= second <= 59

    @staticmethod
    def isValidOffset(secondsOffset: INTEGER) -> bool:
        if secondsOffset is None:
            return True
        return -18 * 3600 <= secondsOffset < 18 * 3600



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import re
from zoneinfo import ZoneInfo

from datetime import datetime, date, time
from typing import Optional, Any

from lxml import etree
import elementpath

from jdmn.feel.lib.Types import STRING, BOOLEAN, INTEGER, LIST, DECIMAL, TIME_OR_DATE_TIME


class DefaultStringLib:
    def string(self, from_: Any) -> STRING:
        if isinstance(from_, datetime):
            return self.print(from_)
        elif isinstance(from_, date):
            return from_.isoformat()
        elif isinstance(from_, time):
            return self.print(from_)

        return str(from_)

    def print(self, from_: TIME_OR_DATE_TIME) -> str:
        tz = from_.tzinfo
        if isinstance(tz, ZoneInfo):
            copy = from_
            copy = copy.replace(tzinfo=None)
            isoformat = copy.isoformat()
            return "{}@{}".format(isoformat, str(tz))
        else:
            return from_.isoformat()

    def contains(self, string: STRING, match: STRING) -> BOOLEAN:
        if string is None or match is None:
            return None

        return match in string

    def startsWith(self, string: Optional[str], match: Optional[str]) -> Optional[bool]:
        if string is None or match is None:
            return None

        return string.startswith(match)

    def endsWith(self, string: STRING, match: STRING) -> BOOLEAN:
        if string is None or match is None:
            return None

        return string.endswith(match)

    def stringLength(self, string: STRING) -> INTEGER:
        if string is None:
            return None

        # The number of Unicode code units in the string
        bytes = string.encode('utf-16', 'surrogatepass')
        # The number of characters (Unicode code point)
        transformed = bytes.decode('utf-16')
        result = len(transformed)
        return result

    def substring(self, string: STRING, startPosition: DECIMAL, length: DECIMAL) -> STRING:
        if string is None or startPosition is None:
            return None

        start = int(startPosition)
        if start < 0:
            start = len(string) + start
        else:
            start -= 1

        normal = self.normalizeSurrogatePairs(string)
        end = len(normal) if length is None else start + int(length)
        result = self.substringCodePoints(normal, start, end)
        return result

    def upperCase(self, string: STRING) -> STRING:
        if string is None:
            return None

        return string.upper()

    def lowerCase(self, string: STRING) -> STRING:
        if string is None:
            return None

        return string.lower()

    def substringBefore(self, string: STRING, match: STRING) -> STRING:
        if string is None or match is None:
            return None

        i = string.find(match)
        return "" if i == -1 else string[0:i]

    def substringAfter(self, string: STRING, match: STRING) -> STRING:
        if string is None or match is None:
            return None

        i = string.find(match)
        return "" if i == -1 else string[i + len(match):]

    def replace(self, input: STRING, pattern: STRING, replacement: STRING, flags: STRING) -> STRING:
        if input is None or pattern is None or replacement is None:
            return None
        if flags is None:
            flags = ""

        return self.evaluateReplace(input, pattern, replacement, flags)

    def matches(self, input: STRING, pattern: STRING, flags: STRING) -> BOOLEAN:
        if input is None or pattern is None:
            return None
        if flags is None:
            flags = ""

        return self.evaluateMatches(input, pattern, flags)

    def split(self, string: STRING, delimiter: STRING) -> LIST:
        if string is None or delimiter is None:
            return None
        if string.strip() == "" or delimiter.strip() == "":
            return None

        result = []
        pattern = re.compile(delimiter)
        start = 0
        for match in re.finditer(pattern, string):
            delimiterStart = match.start()
            delimiterEnd = match.end()
            token = string[start:delimiterStart]
            result.append(token)
            start = delimiterEnd
        if start <= len(string):
            token = string[start:]
            result.append(token)
        return result

    def evaluateReplace(self, input: str, pattern: str, replacement: str, flags: str) -> str:
        expression = "replace(/root, '{}', '{}', '{}')".format(pattern, replacement, flags)
        return self.evaluateXPath(input, expression)

    def evaluateMatches(self, input: str, pattern: str, flags: str) -> bool:
        expression = "/root[matches(., '{}', '{}')]".format(pattern, flags)
        value = self.evaluateXPath(input, expression)
        return len(value) != 0

    @staticmethod
    def evaluateXPath(input: str, expression: str):
        # Read document
        xmlStr = "<root>" + input + "</root>"
        root = etree.fromstring(xmlStr)

        # Evaluate xpath
        res1 = elementpath.select(root, expression)
        return res1

    @staticmethod
    def normalizeSurrogatePairs(string):
        if string is None:
            return None

        # Roundtrip to utf-16 to replace surrogate pairs
        utf16bytes = string.encode('utf-16', 'surrogatepass')
        utf16chars = utf16bytes.decode('utf-16')
        utf8 = utf16chars.encode('utf-8').decode('utf-8')
        return utf8

    @staticmethod
    def substringCodePoints(string, start: int, end: int) -> str:
        result = ""
        cps = [ord(c) for c in string]
        for i, v in enumerate(cps):
            if start <= i < end:
                result += chr(v)
        return result



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional

from jdmn.feel.lib.Utils import varArgToList
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.TernaryBooleanLogicUtil import TernaryBooleanLogicUtil


class DefaultBooleanType(BaseType):
    def __init__(self):
        BaseType.__init__(self)

    def isBoolean(self, value: Any) -> bool:
        return isinstance(value, bool)

    def booleanValue(self, value: Optional[bool]) -> Optional[bool]:
        return value

    def booleanNot(self, operand: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().not_(operand)

    def booleanOr(self, *args) -> Optional[bool]:
        operands = varArgToList(*args)

        if len(operands) < 2:
            return None
        result = operands[0]
        for opd in operands[1:]:
            result = self.binaryBooleanOr(result, opd)
        return result

    def binaryBooleanOr(self, first: Any, second: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().or_(first, second)

    def booleanAnd(self, *args) -> Optional[bool]:
        operands = varArgToList(*args)

        if len(operands) < 2:
            return None
        result = operands[0]
        for opd in operands[1:]:
            result = self.binaryBooleanAnd(result, opd)
        return result

    def binaryBooleanAnd(self, first: Any, second: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().and_(first, second)

    def booleanIs(self, first: Optional[bool], second: Optional[bool]) -> Optional[bool]:
        return self.booleanEqual(first, second)

    def booleanEqual(self, first: Optional[bool], second: Optional[bool]) -> Optional[bool]:
        if (first is None and second is None):
            return True
        elif first is None:
            return False
        elif second is None:
            return False
        else:
            return self.isBoolean(first) and self.isBoolean(second) and first == second

    def booleanNotEqual(self, first: Optional[bool], second: Optional[bool]) -> Optional[bool]:
        return self.booleanNot(self.booleanEqual(first, second))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any


class Arguments(dict):
    def __str__(self):
        return ", ".join(["{0}= '{1}'".format(key, value) for key, value in self.items()])

    def get(self, key: str) -> Any:
        return self[key]

    def put(self, key: str, value: Any) -> None:
        self[key] = value



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import date
from typing import Optional

from isodate import Duration

from jdmn.feel.lib.type.time.DefaultCalendarType import DefaultCalendarType
from jdmn.feel.lib.type.time.DefaultDateComparator import DefaultDateComparator


class DefaultDateType(DefaultCalendarType):
    def __init__(self):
        super().__init__()
        self.__dateComparator = DefaultDateComparator()

    def dateIs(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        if first is None or second is None:
            return first == second

        return self.sameDate(first, second)

    def dateEqual(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.__dateComparator.equalTo(first, second)

    def dateNotEqual(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.booleanType.booleanNot(self.dateEqual(first, second))

    def dateLessThan(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.__dateComparator.lessThan(first, second)

    def dateGreaterThan(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.__dateComparator.greaterThan(first, second)

    def dateLessEqualThan(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.__dateComparator.lessEqualThan(first, second)

    def dateGreaterEqualThan(self, first: Optional[date], second: Optional[date]) -> Optional[bool]:
        return self.__dateComparator.greaterEqualThan(first, second)

    def dateSubtract(self, first: Optional[date], second: Optional[date]) -> Optional[Duration]:
        if first is None or second is None:
            return None

        # Implicit date time conversion
        if self.isDate(first):
            first = self.dateToDateTime(first)
        if self.isDate(second):
            second = self.dateToDateTime(second)
        if self.canNotSubtract(first, second):
            return None

        durationInSeconds = self.dateTimeValue(first) - self.dateTimeValue(second)
        return Duration(seconds=durationInSeconds)

    def dateAddDuration(self, date_: Optional[date], duration_) -> Optional[date]:
        if date_ is None or duration_ is None:
            return None

        return date_ + duration_

    def dateSubtractDuration(self, date_: Optional[date], duration_: Optional[Duration]) -> Optional[date]:
        if date_ is None or duration_ is None:
            return None

        return self.dateAddDuration(date_, - duration_)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Set, List, Union

from jdmn.runtime.annotation.Annotation import Annotation


class AnnotationSet(list):
    def addAnnotation(self, decisionName: str, ruleIndex: int, annotation: Union[str, List[str]]) -> None:
        if isinstance(annotation, list):
            if not (annotation is None) and not len(annotation) == 0:
                annotation = " ".join(annotation)
                self.addAnnotation(decisionName, ruleIndex, annotation)
        elif annotation:
            # Rules index starts from 0
            element = Annotation(decisionName, ruleIndex + 1, annotation)
            self.append(element)

    def toSet(self) -> Set[Annotation]:
        return set(self)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from enum import Enum


class DRGElementKind(Enum):
    DECISION = "decision"
    DECISION_SERVICE = "decision service"
    BUSINESS_KNOWLEDGE_MODEL = "bkm"
    KNOWLEDGE_SOURCE = "knowledge source"
    INPUT_DATA = "input"
    OTHER = "other"

    def __new__(cls, *args, **kwds):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, name):
        self.displayName = name

    def fromName(self, name: str) -> 'DRGElementKind':
        for policy in DRGElementKind:
            if policy.displayName == name:
                return policy
        return self.OTHER



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from enum import Enum


class NumericRoundingMode(Enum):
    UP = "up", "ROUND_UP",
    DOWN = "down", "ROUND_DOWN",
    HALF_UP = "half up", "ROUND_HALF_UP"
    HALF_DOWN = "half down", "ROUND_HALF_DOWN"
    HALF_EVEN = "half even", "ROUND_HALF_EVEN",
    ROUND_FLOOR = "round floor", "ROUND_FLOOR"
    ROUND_CEILING = "round floor", "ROUND_CEILING"
    UNKNOWN = "", ""

    def __new__(cls, *args, **kwds):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, name, pMode):
        self.displayName = name
        self.pMode = pMode

    @staticmethod
    def fromName(name: str) -> 'NumericRoundingMode':
        for mode in NumericRoundingMode:
            if mode.displayName == name:
                return mode
        return NumericRoundingMode.UNKNOWN

    def allowedValues(self):
        return [x for x in NumericRoundingMode if x != NumericRoundingMode.UNKNOWN]



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import datetime


class BaseType:
    UTC = datetime.timezone.utc
    EP_YEAR = 1970
    EP_MONTH = 1
    EP_DAY = 1
    EP = datetime.datetime(EP_YEAR, EP_MONTH, EP_DAY, 0, 0, 0, tzinfo=UTC)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from decimal import Decimal

from jdmn.feel.lib.BaseStandardFEELLib import BaseStandardFEELLib
from jdmn.feel.lib.Types import INTEGER, DECIMAL


class DefaultStandardFEELLib(BaseStandardFEELLib):
    def __init__(self):
        BaseStandardFEELLib.__init__(self)

    #
    # Extra conversion functions
    #
    def valueOf(self, number: INTEGER) -> DECIMAL:
        if number is None:
            return None
        else:
            return Decimal(number)

    def intValue(self, number: DECIMAL) -> INTEGER:
        if number is None:
            return None
        else:
            return int(number)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional, List, Callable

from jdmn.feel.lib.type.bool.TernaryBooleanLogicUtil import TernaryBooleanLogicUtil


class EqualityComparator:
    def __init__(self):
        pass

    def equalTo(self, first: Any, second: Any) -> Optional[bool]:
        pass

    def notEqualTo(self, first: Any, second: Any) -> Optional[bool]:
        return TernaryBooleanLogicUtil().not_(self.equalTo(first, second))

    @staticmethod
    def applyOperator(first: Any, second: Any, result: List[Callable]) -> Optional[bool]:
        if first is None and second is None:
            return result[0]()
        elif first is None:
            return result[1]()
        elif second is None:
            return result[2]()
        else:
            return result[3]()



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional

from jdmn.feel.lib.Types import STRING
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.string.StringComparator import StringComparator


class DefaultStringType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.stringComparator = StringComparator()

    def isString(self, value: Any) -> bool:
        return isinstance(value, str)

    def stringValue(self, value: STRING) -> STRING:
        return value

    def stringIs(self, first: STRING, second: STRING) -> Optional[bool]:
        return first is second

    def stringEqual(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.equalTo(first, second)

    def stringNotEqual(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.notEqualTo(first, second)

    def stringLessThan(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.lessThan(first, second)

    def stringGreaterThan(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.greaterThan(first, second)

    def stringLessEqualThan(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.lessEqualThan(first, second)

    def stringGreaterEqualThan(self, first: STRING, second: STRING) -> Optional[bool]:
        return self.stringComparator.greaterEqualThan(first, second)

    def stringAdd(self, first: STRING, second: STRING) -> STRING:
        if first is None or second is None:
            return None
        else:
            return first + second



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License")or you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.feel.lib.Types import BOOLEAN
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType


class DefaultFunctionType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.booleanType = DefaultBooleanType()

    def isFunction(self, value: Any) -> bool:
        if value is None:
            return True

        raise Exception("Not supported yet")

    def functionValue(self, value: Any) -> Any:
        if (self.isFunction(value)):
            return value
        else:
            return None

    def functionIs(self, function1: Any, function2: Any) -> BOOLEAN:
        return self.functionEqual(function1, function2)

    def functionEqual(self, function1: Any, function2: Any) -> BOOLEAN:
        if self.isFunction(function1) and self.isFunction(function2):
            return function1 == function2
        else:
            return None

    def functionNotEqual(self, function1: Any, function2: Any) -> BOOLEAN:
        return self.booleanType.booleanNot(self.functionEqual(function1, function2))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import logging
from typing import Any

from jdmn.runtime.cache.Cache import Cache


class DefaultCache(Cache):
    LOGGER = logging.getLogger(__name__)

    def __init__(self):
        self.bindings = {}

    def contains(self, key: str) -> bool:
        return key in self.bindings.keys()

    def bind(self, key: str, value: Any) -> None:
        self.LOGGER.debug("Bind '%s' to '%s'", key, value)
        self.bindings[key] = value

    def lookup(self, key: str) -> Any:
        value = self.bindings[key]

        self.LOGGER.debug("Retrieve '%s' = '%s'", key, value)

        return value

    def clear(self) -> None:
        self.bindings.clear()



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.runtime.listener.Arguments import Arguments
from jdmn.runtime.listener.DRGElement import DRGElement
from jdmn.runtime.listener.Rule import Rule


class EventListener:
    def startDRGElement(self, element: DRGElement, arguments: Arguments) -> None:
        pass

    def endDRGElement(self, element: DRGElement, arguments: Arguments, output: Any, duration: int) -> None:
        pass

    def startRule(self, element: DRGElement, rule: Rule) -> None:
        pass

    def matchRule(self, element: DRGElement, rule: Rule) -> None:
        pass

    def endRule(self, element: DRGElement, rule: Rule, result: Any) -> None:
        pass

    def matchColumn(self, rule: Rule, columnIndex: int, result: Any) -> None:
        pass



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import typing


class Annotation:
    # Index starts from 1

    def __init__(self, decisionName: str, ruleIndex: int, annotation: str):
        self.decisionName = decisionName
        self.ruleIndex = ruleIndex
        self.annotation = annotation

    def __eq__(self, other: typing.Any) -> bool:
        if self is other:
            return True
        if type(self) is not type(other):
            return False

        if self.decisionName != other.decisionName:
            return False
        if self.ruleIndex != other.ruleIndex:
            return False
        if self.annotation != other.annotation:
            return False

        return True

    def __str__(self) -> str:
        return "Annotation('{0}', {1}, '{2}')".format(self.decisionName, self.ruleIndex, self.annotation)
